/*
 * Copyright (c) 2003, Raymond Bosman
 * Copyright (c) 2003, Frederik Holljen
 * All Rights Reserved.
 *
 * See COPYING for licensing.
 */


header "post_include_hpp"
{
#include "../antlr/CkASTNodeBase.hpp"

#include "../tools/datatype.hpp"
#include "../classencoder/class.hpp"
#include "../classencoder/bytecode.hpp"
#include "../classencoder/method_info.hpp"
#include "../tools/symboltable.hpp"

// ugly defines until namespace is implemented
#define Undefined CkDataType::Undefined
#define PrimVoid CkDataType::PrimVoid
#define PrimBool CkDataType::PrimBool
#define PrimChar CkDataType::PrimChar
#define PrimInt CkDataType::PrimInt
#define Custom CkDataType::Custom

// using clauses in the header is not pretty, but antlr requires it :-(
    using namespace ClassEncoder;
}

header "post_include_cpp"
{
	#include <iostream>
	#include <sstream>
	#include "../tools/errorhandler.hpp"
    #include "../tools/exception.hpp"

	using namespace std;
}

// Global options
options
{
	mangleLiteralPrefix = "TK_"; // instead of TOKEN
	language="Cpp";
}

/**
 * \brief Codegenerator class
 * This class is a treewalker that generates JVM bytecode. Note that it cannot be run before
 * the Checker since it requires some tree decorations.
 */
class CkCodeGenerator extends TreeParser;
options
{
	importVocab=Ck;
    ASTLabelType = "RefCkASTNodeBase";
}

//! Define stuff we need in the class header
{
private:
    Class *TargetClass;
    method_info *CurrentMethod;
    ByteCode *Target;
    SymbolTable<CGIdEntry> Symbol_table;
    SymbolTable<CGFunctionEntry> Function_table;

public:
    /**
     * \brief Set the target class.
     * \param targetClass The Class object that should write the output. This must be set prior
     * to walking the AST tree.
     */
    void setClass( Class *targetClass )
    {
        TargetClass = targetClass;
        // constructor
        method_info *method = TargetClass->addMethod( "<init>" );
        ByteCode *code = method->code();
        code->emit_aload( 0 );
        code->emit_invokespecial( "java/lang/Object", "<init>", "()V" );
        code->emit_return();
        method->finish();
    }
}

/**
 * \brief Top node in the AST tree. Starts treewalking by visiting all the function declarations.
 */
program
   : #(PROGRAM (parseFunction)*)
   ;

/**
 * \brief Called from program, parses a function declaration.
 * This function has the following tasks:
 * - open a scope for the function parameters (and close it after parsing the block)
 * - create the JVM identification string and store this in the Function_table when registering the function.
 */
parseFunction
	: #(FUNCTION_DECL
    {
		Symbol_table.openScope();
        string functionParams;
        string functionRet;
        string functionName;
	}
    p:primitive
    {
        // fetch return type here
        switch( p->getType() )
        {
            case PRIM_VOID: functionRet = "V"; break;
            case PRIM_INT:
            case PRIM_BOOLEAN: functionRet = "I"; break;
            case PRIM_CHAR: functionRet = "C"; break;
            default:
                throw Exception( __FILE__, __LINE__, "Wrong datatype" );
                break;
        }
    }
        (id:IDENT // normal function
    {
        CurrentMethod = TargetClass->addMethod( id->getText() );
        functionName = id->getText();
        CurrentMethod->setAccess( ACC_PUBLIC | ACC_STATIC );
        Target = CurrentMethod->code();
    }
    (par:parameter
    {
        string param = par->getResultType().toJVMString();
        functionParams += param;
        CurrentMethod->addParameter( param );
    }
    )* | MAIN // main is special since it requires params in java
    {
        CurrentMethod = TargetClass->addMethod( "main" );
        CurrentMethod->setAccess( ACC_PUBLIC | ACC_STATIC );
        CurrentMethod->addParameter( "[Ljava/lang/String;", true );

        functionName = "main";
        functionParams += "[Ljava/langString;";
        // cheat!
        functionRet = "V";
        Target = CurrentMethod->code();
    }
    )
    {
        // register function, this must happen before block in order to make recursion work
        CGFunctionEntry *fEntry = new CGFunctionEntry;
        fEntry->FunctionID = "(" + functionParams + ")" + functionRet;
        if( functionRet != "V" )
            fEntry->Returns = true;
        Function_table.enter( functionName, fEntry );
    }
        block[true])
    {
        // close scope
        Symbol_table.closeScope();

        // set correct return type for function
        CurrentMethod->setReturnType( functionRet );

        // emit final return
        if( functionRet != "V" )
            Target->emit_ireturn();
        else
            Target->emit_return();
        CurrentMethod->finish();
    }
	;

/**
 * \brief Parse a function parameter declaration.
 * registers the parameter as local var for the function and enter it in the Symbol_table
 */
parameter
    : #(PARAM_DEF (CONST)? primitive (id:IDENT
	{
		try
        {
            unsigned int localPos = 0;
            localPos = CurrentMethod->addLocal();
            Symbol_table.enter( id->getText(), new CGIdEntry( localPos ) );
        }catch( SymbolTableException & )
        {
            throw Exception( __FILE__, __LINE__, "Variable unknown" );
        }
	}

	| #(arrayId:ARRAY IDENT
	{
	    unsigned int localPos = 0;
        try
        {
            localPos = CurrentMethod->addLocal();
            Symbol_table.enter( arrayId->getText(), new CGIdEntry( localPos ) );
        }catch( SymbolTableException & )
        {
            throw Exception( __FILE__, __LINE__, "Variable already declared" );
        }
	}

	)))
	;


/**
 * \brief Parse a block ( code in { } ).
 * \param resultPush If this is false, the last expression won't push a result to the stack.
 * Tracks all the declarations and expressions in the block. None of the expressions should not produce a
 * result except the last one if resultPush was specified.
 */
block
[bool resultPush]
    : #(b:BLOCK
    {
        bool getRet = false;
        antlr::RefAST next = b->getFirstChild();
        if( resultPush && next && next->getNextSibling() == 0 ) getRet = true;
        if( !b->getIgnoreScope() )
            Symbol_table.openScope();
    }
        ((declaration | ex:expression[getRet] | ret:returnExpr[false])
        {
            if( resultPush )
            {
                next = next->getNextSibling();
                if( next && next->getNextSibling() == 0 ) getRet = true;
            }
        }
     )*)
    {
        try
        {
            if( !b->getIgnoreScope() )
                Symbol_table.closeScope();
        }catch( SymbolTableException & )
        {
            ErrorHandler::error( b->getLine(), "Compiler oops: Unmatched scopes @ CkChecker::block" );
        }
    }
	;

/**
 * \brief Parses a return expression.
 * \param resultPush Should the return actually produce a result..
 */
returnExpr
[bool resultPush]
    : #(RETURN expression[true])
	;


/**
 * \brief Parse the common parts of a variable declaration.
 * Parses the parts all variable declarations have in common and starts the parsing of the
 * optional initialisers. If everything goes right it retrieves a new variable position from
 * the classencoder and stores it in the symbol table.
 */
declaration
    : #(DECLARATION (CONST)? prim:primitive id:varDeclaration
                    {
                        unsigned int localPos;
                        if( Symbol_table.retrieve( id->getText() ) == 0 )
                        {
                            throw Exception( __FILE__, __LINE__, "Variable already declared" );
                        }
                        else
                        {
                            localPos = Symbol_table.retrieve( id->getText() )->JVMId;
                        }
                    }
                    (expr:expression[true])? (arrayInit[localPos])? )
    {
        if( expr  )
        {
            Target->emit_istore( localPos );
        }
    }
	;

/**
 * \brief Parses the specific parts of an array declaration and a normal declaration.
 * - For the normal declarations the variable is simply added to the symbol table.
 * - For an array the variable is added to the symbol table. The type and size(s) of the array
 * is worked out and code to create the array is emited.
 */
varDeclaration
    : id:IDENT // normal declaration, not much to to except register ourselves
    {
        try
        {
            unsigned int localPos = 0;
            localPos = CurrentMethod->addLocal();
            Symbol_table.enter( id->getText(), new CGIdEntry( localPos ) );
        }catch( SymbolTableException & )
        {
            throw Exception( __FILE__, __LINE__, "Variable unknown" );
        }
    }
    | #(array:ARRAY arrayId:IDENT d1:INT (d2:INT)? )
    // array declaration, create the array
    {
        unsigned int localPos = 0;
        try
        {
            localPos = CurrentMethod->addLocal();
            Symbol_table.enter( arrayId->getText(), new CGIdEntry( localPos ) );
        }catch( SymbolTableException & )
        {
            throw Exception( __FILE__, __LINE__, "Variable already declared" );
        }

        stringstream ss( d1->getText() );
        int dim1;
        ss >> dim1;

        if( d2 ) // two dimensions
        {
            Target->emit_bipush( dim1 );
            stringstream ss( d2->getText() );
            int dim2;
            ss >> dim2;
            Target->emit_bipush( dim2 );
            switch( array->getResultType().getReturnType() )
            {
                case PrimBool:
                    Target->emit_multianewarray( "[[I", 2 );
                    break;
                case PrimChar:
                    Target->emit_multianewarray( "[[C", 2 );
                    break;
                case PrimInt:
                    Target->emit_multianewarray( "[[I", 2 );
                    break;
                default:
                    throw Exception( __FILE__, __LINE__, "Wrong result type" );
                    break;
            }
            Target->emit_astore( localPos );
        }
        else    // One Dimension
        {
            ClassEncoder::_basic_data_types type = ClassEncoder::T_BOOLEAN;
            switch( array->getResultType().getReturnType() )
            {
                case PrimBool: type = ClassEncoder::T_INT; break;
                case PrimChar: type = ClassEncoder::T_CHAR; break;
                case PrimInt: type = ClassEncoder::T_INT; break;
                default:
                    throw Exception( __FILE__, __LINE__, "Unknown type" );
                    break;
            }
            Target->emit_bipush( dim1 );
            Target->emit_newarray( type );
            Target->emit_astore( localPos );
        }
    }
    ;

/**
 * \brief Parse an array initialisation for either single or double arrays.
 * \param localPos The JVMId of the variable to write to.
 * Walks the initialisation tree's like a for loop, and simply stores each entry along the way.
 */
arrayInit
[unsigned int localPos]
    : #(si:SINGLE_ARRAY_INIT { int count = -1;} (
        {  // run before each operand
            count++;
            Target->emit_aload( localPos );
            Target->emit_sipush( count );
        }
        op:operand[true]
        { // run after each operand
            switch( si->getResultType().getReturnType() )
            {
                case PrimBool: Target->emit_iastore(); break;
                case PrimChar: Target->emit_castore();  break;
                case PrimInt: Target->emit_iastore(); break;
                default:
                    throw Exception( __FILE__, __LINE__, "Unknown type" );
                    break;
            }
        }
        )+)
    | #(di:DOUBLE_ARRAY_INIT { int dim1Count = -1;}
        ({ dim1Count++;} #(SINGLE_ARRAY_INIT { int dim2Count = -1; }(
                               {
                                   dim2Count++;
                                   Target->emit_aload( localPos );
                                   Target->emit_sipush( dim1Count );
                                   Target->emit_aaload( );
                                   Target->emit_sipush( dim2Count );
                               }sop:operand[true]
                               {
                                   switch( di->getResultType().getReturnType() )
                                   {
                                       case PrimBool: Target->emit_iastore(); break;
                                       case PrimChar: Target->emit_castore();  break;
                                       case PrimInt: Target->emit_iastore(); break;
                                       default:
                                           throw Exception( __FILE__, __LINE__, "Unknown type" );
                                           break;
            }
                               }
                               )+))+ )
    ;

/**
 * \brief The main parsing method for expressions.
 * \param resultPush set this if you want the expression to push its result onto the stack.
 * Propagates the value of resultPush to sub parsing functions if needed.
 */
expression
[bool resultPush]
	/* highest priority */
	: assign[resultPush]

	/* priority 6 */
/*!
 * \brief || operation
 * Emits code that leaves a 1 on the stack for true, 0 for false.
 */
    | #(orOp:OR lh:expression[resultPush]
		{
            unsigned int orAddr;
			if( resultPush )
                orAddr = Target->emit_ifne(0);
		}
		rh:expression[resultPush])
		{
            if( resultPush )
            {
                // Patch the lh expression. Jump to the end.
                Target->patch (orAddr, Target->nextInstrAddr() + 7 - orAddr);

                // Both types are bool.
                Target->emit_ifne(7); 	// (3 bytes) Jump to iconst(1)
                Target->emit_iconst(0); 	// (1 byte)
                Target->emit_goto(4);  		// (3 bytes) Jump to end
                Target->emit_iconst(1); 	// (1 byte)
            }
		}

	/* priority 5 */
/*!
 * \brief && operation
 * Emits code that leaves a 1 on the stack for true, 0 for false.
 */
	| #(andOp:AND andlh:expression[resultPush]
		{
            unsigned int andAddr;
			if( resultPush )
                andAddr = Target->emit_ifeq(0);
		}

	andrh:expression[resultPush]) /* ** AND ** */
    {
        if( resultPush )
        {
            // patch the previous expression.
            Target->patch (andAddr, Target->nextInstrAddr() + 7 - andAddr);

            // Both types are bool.
            Target->emit_ifeq(7); 	// (3 bytes) Jump to iconst(0)
            Target->emit_iconst(1); 	// (1 byte)
            Target->emit_goto(4);  		// (3 bytes) Jump to end
            Target->emit_iconst(0); 	// (1 byte)
        }
    }

	/* priority 4 */
/*!
 * \brief < operation
 * Emits code that leaves a 1 on the stack for true, 0 for false.
 */
    | #(ltOp:LTHEN ltlh:expression[resultPush] ltrh:expression[resultPush]) /* ** LT ** */
    {
        if( resultPush )
        {
            Target->emit_if_icmplt(7); 	// (3 bytes) Jump to iconst(0)
			Target->emit_iconst(0); 	// (1 byte)
			Target->emit_goto(4);  		// (3 bytes) Jump to end
			Target->emit_iconst(1); 	// (1 byte)
        }
	 }

/*!
 * \brief <= operation
 * Emits code that leaves a 1 on the stack for true, 0 for false.
 */
	| #(leOp:LEQUAL lelh:expression[resultPush] lerh:expression[resultPush]) /* ** LEQUAL ** */
    {
        if( resultPush )
        {
   			Target->emit_if_icmple(7); 	// (3 bytes) Jump to iconst(0)
			Target->emit_iconst(0); 	// (1 byte)
			Target->emit_goto(4);  		// (3 bytes) Jump to end
			Target->emit_iconst(1); 	// (1 byte)
        }
	}

/*!
 * \brief > operation
 * Emits code that leaves a 1 on the stack for true, 0 for false.
 */
    | #(gtOp:GTHEN gtlh:expression[resultPush] gtrh:expression[resultPush]) /* ** GTHEN ** */
    {
 		if( resultPush )
        {
            Target->emit_if_icmpgt(7); 	// (3 bytes) Jump to iconst(0)
            Target->emit_iconst(0); 	// (1 byte)
            Target->emit_goto(4);  		// (3 bytes) Jump to end
            Target->emit_iconst(1); 	// (1 byte)
        }
    }

/*!
 * \brief >= operation
 * Emits code that leaves a 1 on the stack for true, 0 for false.
 */
    | #(geOp:GEQUAL gelh:expression[resultPush] gerh:expression[resultPush]) /* ** GEQUAL ** */
    {
 		if( resultPush )
        {
            Target->emit_if_icmpge(7); 	// (3 bytes) Jump to iconst(0)
            Target->emit_iconst(0); 	// (1 byte)
            Target->emit_goto(4);  		// (3 bytes) Jump to end
            Target->emit_iconst(1); 	// (1 byte)
        }
    }

/*!
 * \brief == operation
 * Emits code that leaves a 1 on the stack for true, 0 for false. Since this operation supports
 * array's these are treated seperatly. The actuall comparing code for arrays is in libCk.
 */
    | #(eqOp:EQUAL eqlh:expression[resultPush] eqrh:expression[resultPush]) /* ** EQUAL ** */
    {
 		if( resultPush )
        {
            if( eqlh->getResultType().getNumDimensions() == 0 )
            {
                Target->emit_if_icmpeq(7); 	// (3 bytes) Jump to iconst(0)
                Target->emit_iconst(0); 	// (1 byte)
                Target->emit_goto(4);  		// (3 bytes) Jump to end
                Target->emit_iconst(1); 	// (1 byte)
            }
            else if( eqlh->getResultType().getNumDimensions() == 1 )
            {
                Target->emit_sipush( eqlh->getResultType().getDim1Size() );
                if( eqlh->getResultType().getReturnType() == PrimChar )
                    Target->emit_invokestatic( "libCk", "sameCharSingle" ,"([C[CI)I");
                else
                    Target->emit_invokestatic( "libCk", "sameIntSingle" ,"([I[II)I");
            }
            else if( eqlh->getResultType().getNumDimensions() == 2 )
            {
                Target->emit_sipush( eqlh->getResultType().getDim1Size() );
                Target->emit_sipush( eqlh->getResultType().getDim2Size() );

                if( eqlh->getResultType().getReturnType() == PrimChar )
                    Target->emit_invokestatic( "libCk", "sameCharDouble" ,"([[C[[CII)I");
                else
                    Target->emit_invokestatic( "libCk", "sameIntDouble" ,"([[I[[III)I");
            }
        }
    }

/*!
 * \brief != operation
 * Emits code that leaves a 1 on the stack for true, 0 for false. Since this operation supports
 * array's these are treated seperatly. The actuall comparing code for arrays is in libCk.
 */
    | #(neOp:NEQUAL nelh:expression[resultPush] nerh:expression[resultPush]) /* ** NEQUAL ** */
    {
        if( resultPush )
        {
            if( nelh->getResultType().getNumDimensions() == 0 )
            {
                Target->emit_if_icmpne(7); 	// (3 bytes) Jump to iconst(0)
                Target->emit_iconst(0); 	// (1 byte)
                Target->emit_goto(4);  		// (3 bytes) Jump to end
                Target->emit_iconst(1); 	// (1 byte)
            }
            else if( nelh->getResultType().getNumDimensions() == 1 )
            {
                Target->emit_sipush( nelh->getResultType().getDim1Size() );
                if( nelh->getResultType().getReturnType() == PrimChar )
                    Target->emit_invokestatic( "libCk", "nSameCharSingle" ,"([C[CI)I");
                else
                    Target->emit_invokestatic( "libCk", "nSameIntSingle" ,"([I[II)I");
            }
            else if( nelh->getResultType().getNumDimensions() == 2 )
            {
                Target->emit_sipush( nelh->getResultType().getDim1Size() );
                Target->emit_sipush( nelh->getResultType().getDim2Size() );

                if( nelh->getResultType().getReturnType() == PrimChar )
                    Target->emit_invokestatic( "libCk", "nSameCharDouble" ,"([[C[[CII)I");
                else
                    Target->emit_invokestatic( "libCk", "nSameIntDouble" ,"([[I[[III)I");
            }

        }
	 }

	/* priority 3 */ /* * PLUS */
/*!
 * \brief + operation
 * Emit code that adds the two expression pushed on the stack.
 */
    | #(plusOp:PLUS pluslh:expression[resultPush] (plusrh:expression[resultPush]
    {
        if( resultPush )
            Target->emit_iadd();
    }
	)?)

	/* * MINUS * */
/*!
 * \brief - operation
 * Emit code that subtracs the two expression pushed on the stack.
 */
    |	{bool unary = true;}
		#(minusOp:MINUS minuslh:expression[resultPush] (minusrh:expression[resultPush] {unary = false;} )?)
		{
            if (resultPush)
            {
                if (unary)
                    Target->emit_ineg();
                else
                    Target->emit_isub();
            }
		}

	/* priority 2 */
/*!
 * \brief + operation
 * Emit code that multiplies the two expression pushed on the stack.
 */
	| #(multOp:MULTIPLY multlh:expression[resultPush] multrh:expression[resultPush])
    {
		if (resultPush)
        	Target->emit_imul();
    }

/*!
 * \brief / operation
 * Emit code that divides the two expression pushed on the stack.
 */
	| #(divOp:DIVIDE divlh:expression[resultPush] divrh:expression[resultPush])
    {
		if (resultPush)
            Target->emit_idiv();
    }

/*!
 * \brief % operation
 * Emit code that mods the two expression pushed on the stack.
 */
	| #(modOp:MODULO modlh:expression[resultPush] modrh:expression[resultPush])
    {
		if (resultPush)
            Target->emit_irem();
    }
	/* Priority 1 */
/*!
 * \brief ! operation
 * Emit code that reverses the element on the stack. Since this can only be used with boolean
 * expressions a logical xor does the trick. (and is really fast)
 */
    | #(notOp:NOT notExpr:expression[resultPush]) /* ** NOT ** */
    {
        if( resultPush )
        {
            Target->emit_iconst( 1 );
            Target->emit_ixor();
        }
    }

	/* Priority 0 */
	| operand[resultPush]
	| block[resultPush]
	| paren[resultPush]

	/* And expression functions */
	| getStatement[resultPush]
	| putStatement[resultPush]
	| ifStatement[resultPush]
	| whileStatement
	;

/**
 * \brief Handle parentheses by parsing the expression inside of it. Propagate the resultPush.
 */
paren
[bool resultPush]
	: #(PAREN expression[resultPush])
	;

/**
 * \brief Parse an operand which is either a literal, a variable or a function call
 * - literals are simply pushed onto the stack.
 * - function calls generatate a call emit and the result is popped of the stack if noone wants it.
 */
operand
[bool resultPush]
	: variable[resultPush]
	| i:INT
    {
		if (resultPush)
		{
			// push ourselves onto the stack
			stringstream ss( i->getText() );
			int num;
			ss >> num;
			Target->emit_sipush( num );
		}

    }
	| c:CHAR
    {
        if( resultPush )
        {
            // chars are packed in ' ' so char c looks like 'c' hence [1]
            Target->emit_bipush( c->getText()[1] );
        }
    }
	| bt: TRUE
	{
        if( resultPush )
        {
            Target->emit_iconst( 1 );
        }
	}
	| bf: FALSE
	{
        if( resultPush )
        {
            Target->emit_iconst( 0 );
        }
	}
    | #(func:FUNCTION funcID:IDENT 	(paramType:expression[true])*)
    {
        if( Function_table.retrieve( funcID->getText() ) == 0 )
        {
            throw Exception( __FILE__, __LINE__, "Unknown function:" + funcID->getText() );
        }
        else
        {
            string functionID = Function_table.retrieve( funcID->getText() )->FunctionID;
            Target->emit_invokestatic( TargetClass->getClassName(), funcID->getText() , functionID );

            // if noone wants the output don't forget to pop it.
            //! info can be caught from func when it is implemented in the checker
            if( !resultPush && Function_table.retrieve( funcID->getText() )->Returns )
                Target->emit_pop();
        }
    }
	;

/**
 * \brief Parse the usage of a variable. Only used for retreival of the variable value.
 * It is important to notice here that even if the variable call is not an array type the
 * variable can still be an array by definition.  This means that we have to look at the resultType
 * and push correct instructions accordingly.
 *
 * For array calls, we leave the retrieval of the array and it's indexes to arrayGetIndex and only perform
 * the call of the value (either a primitive or an array) here.
 */
variable
[bool resultPush]
	: id:IDENT
    {
        if( resultPush )
        {
            if( Symbol_table.retrieve( id->getText() ) == 0 )
            {
                throw Exception( __FILE__, __LINE__, "Unknown variable: " + id->getText() );
            }
            else
            {
                unsigned int vmindex = Symbol_table.retrieve( id->getText() )->JVMId;
                if( id->getResultType().getNumDimensions() == 0 )
                    Target->emit_iload( vmindex );
                else
                    Target->emit_aload( vmindex );
            }
        }
    }
    | array:arrayGetIndex[resultPush]
    {
        if( resultPush )
        {
            if( array->getResultType().getNumDimensions() > 0) // if array types returns array
            {
                Target->emit_aaload();
            }
            else
            {
                switch( array->getResultType().getReturnType() )
                {
                    case PrimBool: Target->emit_iaload(); break;
                    case PrimChar: Target->emit_caload(); break;
                    case PrimInt:  Target->emit_iaload(); break;
                    default:
                        throw Exception( __FILE__, __LINE__, "Unknown type" );
                        break;
                }
            }
        }
    }
    ;


/**
 * \brief Performs an assign operation.
 * As with variables it is important to notice here that even if the variable call is not an array type the
 * variable can still be an array by definition. This means that we have to look at the resultType
 * and push correct instructions accordingly.
 *
 * - For all non dimensional arrays a simple istore instruction is all that is needed for all basic types.
 * - For array assignment we use functions from libCk.
 *
 */
assign
[bool resultPush]
	:#(op:ASSIGN (lhIdent:IDENT|lhArray:arrayGetIndex[true])
       {
           if( lhArray && lhArray->getResultType().getNumDimensions() != 0 )
               Target->emit_aaload();
       }
       rh:expression[true])
    {
        if( lhIdent )
        {
            if( Symbol_table.retrieve( lhIdent->getText() ) == 0 )
            {
                throw Exception( __FILE__, __LINE__, "Unknown variable: " + lhIdent->getText() );
            }
            unsigned int vmindex = Symbol_table.retrieve( lhIdent->getText() )->JVMId;
            if( lhIdent->getResultType().getNumDimensions() == 0 ) // normal variables
            {
                if( resultPush )
                {
                    Target->emit_dup();
                }
                Target->emit_istore( vmindex );
            }
            else if( lhIdent->getResultType().getNumDimensions() == 1 ) // single array
            {
                Target->emit_aload( vmindex );
                // stack is now ..., source, target
                if( resultPush )
                {
                    Target->emit_dup_x1(); // stack is now ..., target, source, target
                }
                Target->emit_swap();
                Target->emit_sipush( lhIdent->getResultType().getDim1Size() );
                if( lhIdent->getResultType().getReturnType() == PrimChar )
                    Target->emit_invokestatic( "libCk", "singleCharAssign" ,"([C[CI)V");
                else
                    Target->emit_invokestatic( "libCk", "singleIntAssign" ,"([I[II)V");
            }
            else // double array
            {
                Target->emit_aload( vmindex );
                // stack is now ..., source, target
                if( resultPush )
                {
                    Target->emit_dup_x1(); // stack is now ..., target, source, target
                }
                Target->emit_swap();
                Target->emit_sipush( lhIdent->getResultType().getDim1Size() );
                Target->emit_sipush( lhIdent->getResultType().getDim2Size() );
                if( lhIdent->getResultType().getReturnType() == PrimChar )
                    Target->emit_invokestatic( "libCk", "doubleCharAssign" ,"([[C[[CII)V");
                else
                    Target->emit_invokestatic( "libCk", "doubleIntAssign" ,"([[I[[III)V");
            }
        }
        else // array call, top of stack is now: target, index, result
        {
            if( resultPush )
            {
                Target->emit_dup_x2(); // stack is now ...,result, target, index, result
            }
            if( lhArray->getResultType().getNumDimensions() == 0 )
            {
                switch( lhArray->getResultType().getReturnType() )
                {
                    case PrimBool: Target->emit_iastore(); break;
                    case PrimChar: Target->emit_castore();  break;
                    case PrimInt: Target->emit_iastore(); break;
                    default:
                        throw Exception( __FILE__, __LINE__, "Unknown type" );
                        break;
                }
            }
            else // can only be one dimension
            {
                // currently on stack: .., target, source
                if( resultPush )
                {
                    // we want target, target, source
                    Target->emit_swap(); // .., source, target
                    Target->emit_dup_x1(); //.., target, source, target
                    Target->emit_swap(); // ..., target, target, source
                }
                // we want on stack: target, result, dim
                Target->emit_sipush( lhArray->getResultType().getDim1Size() );
                if( lhArray->getResultType().getReturnType() == PrimChar )
                    Target->emit_invokestatic( "libCk", "singleCharAssign" ,"([C[CI)V");
                else
                    Target->emit_invokestatic( "libCk", "singleIntAssign" ,"([I[II)V");
            }
        }
    }
	;

/**
 * \brief Push indexing instructions for an array call. After this you can either get a value
 * or store a value to the given position in the array.
 */
arrayGetIndex[bool resultPush]
    : #(a:ARRAY
    {
        if( Symbol_table.retrieve( a->getText() ) == 0 )
        {
            throw Exception( __FILE__, __LINE__, "Unknown variable: " + a->getText() );
        }
        unsigned int vmindex = Symbol_table.retrieve( a->getText() )->JVMId;

        if( resultPush )
            Target->emit_aload( vmindex );
    }
        // insert aaload of second array
        IDENT expression[resultPush] (
        {
            if( resultPush )
                Target->emit_aaload();
        } expression[resultPush])? )
    ;


/**
 * \brief Parse primitive types.
 */
primitive
	:	"void"
	|	"bool"
	|	"char"
	|	"int"
	;


/**
 * \brief Parse the get statement.
 * \param resultPush Set this if you want the get to push a result (if it can).
 * Emits code that uses java library function calls to retreive values from the console. Different
 * functions are used for different types. The result is stored into the variable.
 * given.
 */
getStatement
[bool resultPush]
	: #(getOp:GET ((id:IDENT|array:arrayGetIndex[true]) {

	// Identifier exist?
	if( id && Symbol_table.retrieve( id->getText() ) == 0 )
    {
        throw Exception( __FILE__, __LINE__, "Unknown variable: " + id->getText() );
    }

    CkDataType::Types basicType = PrimInt;
    if( id )
    {
        basicType = id->getResultType().getReturnType();
    }
    else
    {
        basicType = array->getResultType().getReturnType();
    }

	// Try to read something. Open streams etc.
	// Based on the following java code:
	/* java.io.DataInput in = new java.io.DataInputStream(System.in);
	 * in.readLine());
	 */
	Target->emit_new("java/io/DataInputStream");
	Target->emit_dup();
	Target->emit_getstatic( "java/lang/System", "in", "Ljava/io/InputStream;" );
	Target->emit_invokespecial( "java/io/DataInputStream", "<init>", "(Ljava/io/InputStream;)V" );
	Target->emit_invokeinterface( "java/io/DataInput", "readLine", "()Ljava/lang/String;", 1 );

	if (basicType == PrimChar)
	{
		// Keep the first character from the string.
		Target->emit_iconst(0);
		Target->emit_invokevirtual("java/lang/String", "charAt", "(I)C");
	}
	else if (basicType == PrimInt)
	{
		// Cast to integer.
		Target->emit_invokestatic( "java/lang/Integer", "parseInt", "(Ljava/lang/String;)I" );
	}
	else if (basicType == PrimBool)
	{
		// Read the integer. All numbers result to true, except zero.
		Target->emit_invokestatic( "java/lang/Integer", "parseInt", "(Ljava/lang/String;)I" );

	    Target->emit_ifeq(7); 		// (3 bytes) Jump to iconst(0)
        Target->emit_iconst(1); 	// (1 byte)
        Target->emit_goto(4);  		// (3 bytes) Jump to end
        Target->emit_iconst(0); 	// (1 byte)
	}

    // stack is now the value of the read item.

	// Push result also on the stack?
    if( id )
    {
        if (resultPush && getOp->getResultType() != PrimVoid) Target->emit_dup();

        // And store it.
        Target->emit_istore(Symbol_table.retrieve(id->getText())->JVMId); // (even char)
    }
    else // array... stack is now: arrayref, index, value
    {
        if (resultPush && getOp->getResultType() != PrimVoid) Target->emit_dup_x2();
        Target->emit_iastore();
    }

    } )+ )
	;

/**
 * \brief Parse the put statement.
 * \param resultPush Set this if you want the put to push a result (if it can).
 * Gets the value to display through an expression. Then uses java library calls to
 * display the value on the console.
 */
putStatement
[bool resultPush]
    : #(putOp:PUT (expr:expression[true]
    {
        if( resultPush && putOp->getResultType() != PrimVoid ) // must be single arg, we must dump result
            Target->emit_dup();
        // current situation is the expression to print on top of the stack
        Target->emit_getstatic( "java/lang/System", "out", "Ljava/io/PrintStream;" );
        // the out object should be first.
        Target->emit_swap();

        if( expr->getResultType() == PrimInt || expr->getResultType() == PrimBool)
        {
           Target->emit_invokevirtual( "java/io/PrintStream", "println", "(I)V" );
        }
        else if( expr->getResultType() == PrimChar )
        {
            Target->emit_invokevirtual( "java/io/PrintStream", "println", "(C)V" );
        }
		else
		{
			// Should be a exception.
			ErrorHandler::error( putOp->getLine(),
				"Compiler oops: CkCodeGenerator::putStatement unknown type" );
			exit (4);
		}
    })+ )
	;

/**
 * \brief Parse an if statement.
 * Adds corresponding jump instructions and propagates resultPush if apropriate.
 */
ifStatement
[bool resultPush]
	: #(ifOp:IF expr:expression[true]
    {
        unsigned int toElse = Target->emit_ifeq( 0 );
        unsigned int elsePos = 0;
        unsigned int toEnd = 0;
    }
        ifDo:block[resultPush && ifOp->getResultType() != PrimVoid] (
        {
            toEnd = Target->emit_goto( 0 );
            elsePos = Target->nextInstrAddr();
        }
        elseDo:block[resultPush && ifOp->getResultType() != PrimVoid])?)
        {
            if( elsePos != 0 ) // there was an else part
            {
                // patch if jump to go to else if it failed
                Target->patch( toElse, elsePos - toElse );
                Target->patch( toEnd, Target->nextInstrAddr() - toEnd );
            }
            else
            {
                // no else, else == end;
                Target->patch( toElse, Target->nextInstrAddr() - toElse );
            }
        }
	;

/**
 * \brief Parse a while statement.
 * Adds corresponding jump instructions and propagates resultPush if apropriate.
 */
whileStatement
	: #(whileOp:WHILE
    {
        unsigned int startWhile = Target->nextInstrAddr();
    }
        expr:expression[true]
    {
        unsigned int toEnd = Target->emit_ifeq( 0 );
    }
        whileDo:block[false] )
    {
        Target->emit_goto( startWhile - Target->nextInstrAddr() );
        Target->patch( toEnd, Target->nextInstrAddr() - toEnd );
    }
	;
