/*
 * Copyright (c) 2003, Raymond Bosman
 * Copyright (c) 2003, Frederik Holljen
 * All Rights Reserved.
 *
 * See COPYING for licensing.
 */


header "post_include_hpp"
{
    #include "../antlr/CkASTNodeBase.hpp"
}

header "post_include_cpp"
{
	#include <iostream>
	using namespace std;
}

// Global options.
options
{
	mangleLiteralPrefix = "TK_"; // instead of TOKEN
	language="Cpp";
}

//////////////////////////// PARSER //////////////////////////////////
/**
 * \brief Parser for the Ck language.
 *
 * Checks that the token stream from the lexer is according to the Ck language 
 * specification and builds the corresponding abstract syntax tree.
 */
class CkParser extends Parser ;

options {
    k = 1;                             // # tokens lookahead
    exportVocab = Ck;                 // call the vocabulary "Ck"
    buildAST = true;                    // build an AST
    ASTLabelType = "RefCkASTNodeBase";
	defaultErrorHandler = false;
}

/*!
 * Imaginary tokens that are used for AST construction.
 */
tokens {
	DECLARATION; // primitive with an identifier
    ARRAY; // array primitive
	BLOCK; // The stuff between curly braces
	PAREN; // Stuff between the LPAREN and RPAREN

	PROGRAM;
	FUNCTION;
	FUNCTION_DECL;
	PARAM_DEF;
    SINGLE_ARRAY_INIT; // top node for single array init
    DOUBLE_ARRAY_INIT; // top node for single array init
}


/**
 * \brief program is the toplevel parsing function of Ck.
 *
 * A program can consist of several functions. The functions are encapsulated 
 * in the toplevel PROGRAM node.
 */
program
   : (function)*
 	{#program = #([PROGRAM, "program"], #program); }
   ;

/**
 * \brief function parses a function
 *
 * A function consists of a return type, a function name and several function 
 * parameters. A function is given a FUNCTION_DECL top node.
 */
function
	// If main present, it must be the last function in the program.
	: primitive (IDENT LPAREN! 
		(parameterDef (COMMA! parameterDef)*)? RPAREN! block (SEMI!)*
	| MAIN LPAREN! RPAREN! block (SEMI!)* EOF!)
   { #function = #([FUNCTION_DECL, "function_decl"], #function); }
	;

/**
 * \brief parameterDef parses a function parameter definition.
 *
 * A parameter consists of a an optional "const" declaration a primitive and 
 * a parameterVar. A parameterDef creates a PARAM_DEF top node.
 */
parameterDef
	: (CONST)? primitive parameterVar
   { #parameterDef = #([PARAM_DEF, "param_def"], #parameterDef); }
	;

/**
 * paramterVar parses the identifier and the optional array specification for 
 * a function parameter declaration.
 */
parameterVar
	: id:IDENT (LBRACKET! RBRACKET!
	{
		#parameterVar = #([ARRAY, "array"], #parameterVar);
		//#parameterVar = #([ARRAY, "array"], id, #([INT, "-1"]));
	}
	(LBRACKET RBRACKET!)?
	)?
	;

/**
 * \brief a block is a series of declarations, expression and return 
 * expressions empraced between { and }
 *
 * A block must end with an expression or a return expression.
 */
block
	:	LCURLY! ((declaration (SEMI!)+)* 
			(returnExpr | expression (SEMI!)+))* RCURLY!
		{ #block = #([BLOCK, "block"], #block); }
	;

/**
 * \brief a returnExpr is simply an expression preceeded by the "return" 
 * keyword.
 */
returnExpr
	: RETURN^ expression(SEMI!)+
	;
/**
 * \brief expression is the toplevel parsing function for expressions.
 *
 * The different expression types are handled according to their priority.
 */
expression
	:   orExpression (ASSIGN^ expression)?
    ;

/**
 * \brief Priority 6, the or (||) expression
 */
orExpression
	: andExpression (OR^ andExpression)*
	;


/**
 * \briefPriority 5, the and (&&) expression
 */
andExpression
	: equationExpression (AND^ equationExpression)*
	;

/**
 * \brief Priority 4, the equation expressions.
 *
 * These consists of less than, less or equal than, greater than, greater or 
 * equal than, equal and not equal.
 */
equationExpression
	: plusExpression ((LTHEN^ | LEQUAL^ | GTHEN^ | GEQUAL^ | EQUAL^ | NEQUAL^) 
		plusExpression)*
	;


/**
 * \brief Priority 3, the low priority basic arithmetic.
 *
 * Consists of addition and subtraction.
 */
plusExpression
	: multiplyExpression ((PLUS^ | MINUS^) multiplyExpression)*
	;

/**
 * \brief Priority 2, the high level basic arithmetic
 *
 * Consits of multiplication, division and modulo.
 */
multiplyExpression
    :   unaryExpression ((MULTIPLY^ | DIVIDE^ | MODULO^) unaryExpression)*
    ;

/**
 * \brief Priority 1, unary expressions.
 *
 * Consits of unary + and unary -
 * The unaray expression is the last in the set of operators and is followed 
 * by the language atoms.
 */
unaryExpression
	: (NOT^ | MINUS^ | PLUS^)? atom
	;
/**
 * \brief atom consits of all the expression constructs that are not operators.
 *
 * These are:
 * - literals
 * - variables (either real variables or function calls)
 * - paranteses (with expressions in them)
 * - blocks
 * - atomstatements( get,put,if and while)
 */
atom
	:  INT
	|  CHAR
	|  TRUE | FALSE
	|  variable
	|  paren
	|  block
	|  atomStatements
	;

/**
 * \brief parses variables (also array's) and function calls.
 *
 * Function calls must be handled here since they will always look similar to 
 * variables and we want to keep the language LL(1).
 *
 * If an array is detected the variable is given an ARRAY topnode a normal 
 * identifier keeps the IDENT as topnode.
 * If a function(call) is detected it is given a FUNCTION topnode.
 */
variable
    : IDENT (LBRACKET! expression RBRACKET! (LBRACKET! expression RBRACKET!)? {
            #variable = #([ARRAY, "array"], #variable) ;
	                                            }
    | LPAREN!(expression (COMMA! expression)*)? RPAREN!
            {#variable = #([FUNCTION, "function"], #variable) ; }
												)?
    ;
/**
 * \brief paren is simply an expression within "(" and ")". It is given a 
 * PAREN top node.
 */
paren
	: LPAREN! expression RPAREN!  { #paren = #([PAREN, "paren"], #paren); }
	;

/**
 * \brief atomStatements are function like Ck statements.
 *
 * These consist of the put, get, if and while statements.
 */
atomStatements
	:  getStatement
	|  putStatement
	|  ifStatement
	|  whileStatement
	;

/**
 * \brief parses get statements
 *
 * A get statement looks like a function call but has a variable number of 
 * arguments. All the arguments must be variables.
 * A get statements gets a GET top node.
 */
getStatement
	: (GET! | READ! {
		cout << "'read' function deprecated use 'get' instead" << endl;
					}) LPAREN! variable (COMMA! variable)* RPAREN!
		{
			#getStatement = #([GET, "get"], #getStatement) ;
		}
	;

/**
 * \brief parses put statements
 *
 * put statements are like get statements, except that they accept full 
 * expressions as parameters instead of variables only.
 * A put statements gets a PUT top node.
 */
putStatement
	: (PUT! | PRINT!{
		cout << "'print' function deprecated use 'put' instead" << endl;
					}) LPAREN! expression (COMMA! expression)* RPAREN!
		{
			#putStatement = #([PUT, "put"], #putStatement) ;
		}
	;

// max array size is two dimensions
// can either be an array declaration or a normal declaration
// array dec int a[5] expands to declaration( int array( a 5 ) )
// normald dec int a = 10 expands into declaration( int a 10 )
/**
 * \brief declaration parses a variable declaration.
 *
 * A variable declaration can be either:
 * - a normal variable
 * - a single dimension variable
 * - a double dimension variable
 * All three sorts can have initialisation. For a normal array this is simply 
 * an expression while for arrays this is parsed seperatly.
 * In order to keep things tidy in for the treewalkers we do not use the 
 * automatic tree building mechanism but build the AST tree ourselves.
 * For a normal declaration the tree built looks like:
 * \code
 *               declaration
 *          /      |      |      \
 *     CONST? primitive  IDENT  expression?
 * \endcode
 *
 * For an array declaration tree built looks like:
 * \code
 *                   declaration
 *          /      |      |      |         \
 *     CONST? primitive  IDENT  ARRAY    arrayInit?
 *                              /   \
 *                             INT  INT?
 * \endcode
 */
declaration!
    : (c:CONST)? p:primitive id:IDENT (
            LBRACKET d1:INT RBRACKET (LBRACKET d2:INT RBRACKET)? 
				(ASSIGN! ai:arrayInit)?
    { #declaration = #([DECLARATION, "declaration"], c, p,
      #([ARRAY, "array"], id, d1, d2 ), ai ); 
	}
    | ASSIGN exp:expression
    { #declaration = #([DECLARATION, "declaration"], c, p, id, exp );  }
    | /* empty declaration means normal */
    { #declaration = #([DECLARATION, "declaration"], c, p, id, exp );  }
    )
    ;

/**
 * \brief parses array initialisation
 *
 * Array intialisation is always either a singleArrayInit or a doubleArrayInit 
 * between "{" and "}".
 */
arrayInit
    : ( LCURLY! (singleArrayInit | doubleArrayInit ) RCURLY! )
    ;

/**
 * \brief singleArrayInit is a comma seperated list of literals.
 *
 * A single array initalisation gets the top node SINGLE_ARRAY_INIT.
 */
singleArrayInit
    : ( (INT|CHAR|TRUE|FALSE) (COMMA! (INT|CHAR|TRUE|FALSE))* )
    {
      #singleArrayInit = 
	  	#([SINGLE_ARRAY_INIT, "single_array_init"], #singleArrayInit);
    }
    ;

/**
 * \brief doubleArrayInit is a comma seperated list of singleArrayInit's
 *
 * A double array initalisation gets the top node DOUBLE_ARRAY_INIT.
 */
doubleArrayInit
    : ( (LCURLY! singleArrayInit RCURLY!) 
		( COMMA! LCURLY! singleArrayInit RCURLY!)* )
    {
      #doubleArrayInit = 
	  	#([DOUBLE_ARRAY_INIT, "double_array_init"], #doubleArrayInit) ;
    }
    ;

/**
 * \brief primitive consists of the Ck datatype primitives
 */
primitive
	:	PRIM_VOID
	|	PRIM_BOOLEAN
	|	PRIM_CHAR
	|	PRIM_INT
	;

// Conditional statements
/**
 * \brief ifStatement parses an if statement.
 *
 * An if statement consists of the if keyword followed by an expression 
 * between "(" and ")". After this comes the obligatory block, with the 
 * optional else keyword and it's block after it.
 */
ifStatement
	: IF^ LPAREN! expression RPAREN! block (ELSE! block)?
	;

/**
 * \brief whileStatement parses a while statement.
 *
 * A while statement consits of the while keyword followed by an expression 
 * between "(" and ")" and the obligatory block.
 */
whileStatement
	: WHILE^ LPAREN! expression RPAREN! block
	;

//////////////////////////// LEXER ///////////////////////////////////
/**
 * \brief Lexer for the Ck language.
 */
class CkLexer extends Lexer;
options
{
	k=3; // for /* comment..
	exportVocab=Ck;
	charVocabulary = '\3'..'\377';

    caseSensitive = true;           // lower and upper case is significant
    caseSensitiveLiterals = true;   // literals are case sensitive
    testLiterals = false;           // do not check the tokens table by default
}

/*!
 * These are the keywords of Ck.
 */
tokens
{
	MAIN = "main";
	RETURN = "return";

	READ = "read"; // deprecated
	PRINT = "print"; // deprecated
	GET = "get";
	PUT = "put";

	IF = "if";
	ELSE = "else";
	WHILE = "while";

    CONST = "const";
	// primitives
	PRIM_VOID = "void";
	PRIM_BOOLEAN = "bool";
	PRIM_CHAR = "char";
	PRIM_INT = "int";

	TRUE = "true";
	FALSE = "false";
}



// operators
/** PlUS is denoted by '+' */
PLUS            :   '+' ;
/** MINUS is denoted by '-' */
MINUS           :   '-' ;
/** MULTIPLY is denoted by '*' */
MULTIPLY		:	'*' ;
/** DIVIDE is denoted by '/' */
DIVIDE			:	'/' ;
/** MODULO is denoted by '%' */
MODULO			:	'%' ;
/** COLON is denoted by ';' */
COLON           :   ':' ;
/** SEMI is denoted by ';' */
SEMI			:   ';' ;

/** LPAREN is denoted by '(' */
LPAREN    options{paraphrase="(";} :   '(' ;
/** RPAREN is denoted by ')' */
RPAREN    options{paraphrase=")";} :   ')' ;
/** LBRACKET is denoted by '[' */
LBRACKET  options{paraphrase="[";} :   "[" ;
/** RBRACKET is denoted by ']' */
RBRACKET  options{paraphrase="]";} :   "]" ;

// relational operators

// LT is used in the generated cpp code.
/** LTHEN is denoted by '<' */
LTHEN			: "<";
/** LEQUAL is denoted by "<=" */
LEQUAL			: "<=";
/** GTHEN is denoted by '>' */
GTHEN			: ">";
/** GEQUAL is denoted by ">=" */
GEQUAL			: ">=";
/** EQUAL is denoted by "==" */
EQUAL			: "==";
/** NEQUAL is denoted by "!=" */
NEQUAL			: "!=";

/** AND is denoted by "&&" */
AND			: "&&";
/** OR is denoted by "||" */
OR				: "||";
/** NOT is denoted by "!" */
NOT				: "!";

/** LCURLY is denoted by "{" */
LCURLY			: '{' ;
/** RCURLY is denoted by "}" */
RCURLY			: '}' ;
/** ASSIGN is denoted by "=" */
ASSIGN			: '=' ;
/** COMMA is denoted by "," */
COMMA	        : ',' ;

/**
 * Words out a character
 */
CHAR
	:	'\'' (ESC|~'\'') '\''
	;

/**
 * Words out a string literal
 */
STRING_LITERAL
	:	'"' (ESC|~'"')* '"'
	;

/**
 * Works out escaped string sequences
 */
protected
ESC	:	'\\'
		(	'n'
		|	'r'
		|	't'
		|	'b'
		|	'f'
		|	'"'
		|	'\''
		|	'\\'
		|	'0'..'3'
			(
				options {
					warnWhenFollowAmbig = false;
				}
			:	DIGIT
				(
					options {
						warnWhenFollowAmbig = false;
					}
				:	DIGIT
				)?
			)?
		|	'4'..'7'
			(
				options {
					warnWhenFollowAmbig = false;
				}
			:	DIGIT
			)?
		)
	;

/**
 * And IDENT is a series of characters and digits (always start with a char).
 * Checks for keywords with testLiterals = true.
 */
IDENT
options
{
	testLiterals = true;
	paraphrase = "an identifier";
}
    :   (LOWER | UPPER | SPECIAL) (LOWER | UPPER | DIGIT | SPECIAL)*
	;


// Yeah, protected. Simply: It's not a token which can be recognized.
// It's only a part or subset of an recognizable token. e.g. INT.
/**
 * a..z is a the subset of token e.g. an IDENT
 */
protected
LOWER           :   ('a'..'z') ;

/**
 * A..Z is a the subset of token e.g. an IDENT
 */
protected
UPPER           :   ('A'..'Z') ;

/**
 * 0..9 is a the subset of token e.g. an INT
 */
protected
DIGIT           :   ('0'..'9') ;

/**
 * "_" may be in a identifier and must be handled seperatly since it is 
 * not in a..z
 */
protected
SPECIAL           :  '_' ;


/**
 * An integer is represented by several numbers
 */
INT				:	(DIGIT)+ ;


/**
 * Whitespace - ignore
 */
WS
    :   ( ' '
        |	'\t'
        |	'\f'
        // handle newlines
        |   (	"\r\n"  // Yeah DOS
			|	'\r' 	// Macintosh or DOS
            |	'\n'    // Unix (the right way)
            )
            { newline(); }
        )
        { $setType(antlr::Token::SKIP); }
    ;

/**
 * Single line comment - ignore
 * Single line comments are // like in C++
 */
SL_COMMENT :
	"//"
	(~'\n')* '\n'
	{ _ttype = antlr::Token::SKIP; newline(); }
	;

/**
 * Multi line comment - ignore
 * Multi line comments are like in C/C++
 */
ML_COMMENT
	:	"/*"
		(	{ LA(2)!='/' }? '*'
		|	'\n' { newline(); }
		|	~('*'|'\n')
		)*
		"*/"
		{ $setType(antlr::Token::SKIP); }
	;
